<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 1/26/16
 * Time: 11:25 AM
 */

namespace qfq;

use qfq;
use qfq\Store;
use qfq\UserFormException;

require_once(__DIR__ . '/../../qfq/store/Store.php');
require_once(__DIR__ . '/../../qfq/Constants.php');
require_once(__DIR__ . '/../../qfq/helper/KeyValueStringParser.php');
require_once(__DIR__ . '/../../qfq/exceptions/UserFormException.php');

class HelperFormElement
{

    /**
     * Expand column 'parameter' to row array as virtual columns.
     * E.g.: 'detail=xId:grId\nShowEmptyAtStart=1`
     *
     * @param array $elements
     * @throws \qfq\UserFormException
     */
    public static function explodeParameterInArrayElements(array &$elements) {
        foreach ($elements AS $key => $element) {
            self::explodeParameter($element);
            $elements[$key] = $element;
        }
    }

    /**
     * Take all rows from field $element['parameter'] and merge them with $element. '$element' grows in size.
     *
     * Iterates over all FormElements in $elements.
     * If $element['parameter'] exist and is filled: explode the content and append them to the current array.
     *
     * @param array $elements
     * @throws UserFormException
     */
    public static function explodeParameter(array &$element)
    {
        // Something to explode?
        if (isset($element['parameter']) && $element['parameter'] !== '') {
            // Explode
            $arr = KeyValueStringParser::parse($element['parameter'], "=", "\n");
            // Check if some of the exploded keys conflict with existing keys
            $checkKeys = array_keys($arr);
            foreach ($checkKeys AS $checkKey) {
                if (isset($element[$checkKey])) {
                    $store = Store::getInstance();
                    $store->setVar(SYSTEM_FORM_ELEMENT, Logger::formatFormElementName($element), STORE_SYSTEM);
                    $store->setVar(SYSTEM_FORM_ELEMENT_COLUMN, 'parameter', STORE_SYSTEM);
                    throw new UserFormException("Found reserved keyname '$checkKey'", ERROR_RESERVED_KEY_NAME);
                }
            }
            $element = array_merge($element, $arr);
        }
    }

    /**
     * Build the FE name: <field>:<record index)
     *
     * @param string $field
     * @param string $id
     * @return string
     */
    public static function buildFormElementName($field, $id)
    {
        return ($field . ':' . $id);
    }

    /**
     * Checkboxen, belonging to one element, grouped together by name: <fe>_<field>_<index>
     *
     * @param string $field
     * @param string $index
     * @return string
     */
    public static function prependFormElementIdCheckBoxMulti($field, $index) {
        return ('_' . $index . '_' . $field);
    }

}