<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 5/4/16
 * Time: 1:34 PM
 */

namespace qfq;


class Session {

    private static $instance = null;
    private static $phpUnit = null;
    private static $sessionLocal = array();

    /**
     * @param bool|false $phpUnit
     * @throws CodeException
     */
    private function __construct($phpUnit = false) {
        if (self::$phpUnit !== null)
            throw new CodeException("Try to set flag phpunit again - that should not happen.", ERROR_CODE_SHOULD_NOT_HAPPEN);

        self::$phpUnit = $phpUnit;

        if (self::$phpUnit === true) {
            self::$sessionLocal = array();
        } else {
            session_name(SESSION_NAME);
            session_start();
        }
        self::checkFeUserUid();
    }

    /**
     * Check if the feUserUid is stored in the session (even with 'false' which indicates not logged in user).
     *   If not, clear the session and save the feUser, feUserUid in the session.
     * Check if the recent logged in feUserUid is equal to the one stored in session: If different, invalidate (clear) the session and
     *   save the new feUser, feUserUid in the session.
     * If isset($GLOBALS["TSFE"]), than we're in a T3 environment, else we are called as API classes and need to fake
     *   feUser / feUserUid from previous stored session.
     * It's neccessary to have feUser / feUserUid available in API classes, due to dynamic update which might reload data based on feUser / feUserUid.
     */
    private static function checkFeUserUid() {

        $feUserUidSession = Session::get(SESSION_FE_USER_UID);
        $feUserSession = Session::get(SESSION_FE_USER);

        if (isset($GLOBALS["TSFE"])) {
            // if noone is logged in: 0
            $feUidLoggedIn = isset($GLOBALS["TSFE"]->fe_user->user["uid"]) ? $GLOBALS["TSFE"]->fe_user->user["uid"] : false;
            $feUserSession = isset($GLOBALS["TSFE"]->fe_user->user["username"]) ? $GLOBALS["TSFE"]->fe_user->user["username"] : false;
            $feUserGroup = isset($GLOBALS["TSFE"]->fe_user->user["usergroup"]) ? $GLOBALS["TSFE"]->fe_user->user["usergroup"] : false;
        } else {
            // If we are called through API there is no T3 environment. Assume nothing has changed, and fake the following check to always 'no change'.
            $feUidLoggedIn = $feUserUidSession;
        }

        if ($feUidLoggedIn !== $feUserUidSession) {
            // destroy existing session store
            Session::clear();

            // save new feUserUid, feUserName
            Session::set(SESSION_FE_USER_UID, $feUidLoggedIn);
            Session::set(SESSION_FE_USER, $feUserSession);
            Session::set(SESSION_FE_USER_GROUP, $feUserGroup);

//            throw new UserFormException("FYI: Session has been cleared. Reload this page. ".
//                "feUserUidSession:'$feUserUidSession',  feUserSession:'$feUserSession' isset(TSFE):'" . isset($GLOBALS["TSFE"]) ? 'true' : 'false' );
        }
    }

    /**
     * @param $key
     * @return bool
     */
    public static function get($key) {

        if (self::$phpUnit) {
            if (isset(self::$sessionLocal[$key]))
                $value = self::$sessionLocal[$key];
            else
                $value = false;
        } else {
            if (isset($_SESSION[SESSION_NAME][$key]))
                $value = $_SESSION[SESSION_NAME][$key];
            else
                $value = false;

        }

        return $value;
    }

    /**
     *
     */
    public static function clear() {

        if (self::$phpUnit) {
            self::$sessionLocal = array();
        } else {
            $_SESSION[SESSION_NAME] = array();
        }
    }

    /**
     * @param $key
     * @param $value
     */
    public static function set($key, $value) {

        if (self::$phpUnit) {
            self::$sessionLocal[$key] = $value;
        } else {
            $_SESSION[SESSION_NAME][$key] = $value;
        }
    }

    /**
     * @param bool|false $phpUnit
     * @return Session class
     */
    public static function getInstance($phpUnit = false) {

        // Design Pattern: Singleton
        if (self::$instance === null) {
            self::$instance = new self($phpUnit);
        }

        return self::$instance;
    }

}